'''
base.py: base class definition for archive services.

Authors
-------

Michael Hucka <mhucka@caltech.edu> -- Caltech Library

Copyright
---------

Copyright (c) 2020-2021 by the California Institute of Technology.  This code
is open-source software released under a 3-clause BSD license.  Please see the
file "LICENSE" for more information.
'''


# Class definitions.
# .............................................................................
# Basics for the __eq__ etc. methods came from
# https://stackoverflow.com/questions/1061283/lt-instead-of-cmp

class Service(object):

    label = ''
    name = ''
    color = ''

    def save(self, url):
        '''Send the "url" to the service to save it.'''
        pass


    def newest(self, url):
        '''Return the newest saved version of the content at "url".'''
        pass


    def oldest(self, url):
        '''Return the oldest saved version of the content at "url".'''
        pass


    def nearest(self, url):
        '''Return the closest saved version of the content at "url".'''
        pass


    # The rest of these methods are generic and don't need to be overridden.
    # .........................................................................

    def __str__(self):
        return self.name


    def __repr__(self):
        return self.name


    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        else:
            return not self.name < other.name and not other.name < self.name


    def __ne__(self, other):
        return not __eq__(self, other)


    def __lt__(self, other):
        return self.name < other.name


    def __gt__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        else:
            return other.name < self.name


    def __le__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        else:
            return not other.name < self.name


    def __ge__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        else:
            return not self.name < other.name



'''
exit_codes.py: define exit codes for program return values

Authors
-------

Michael Hucka <mhucka@caltech.edu> -- Caltech Library

Copyright
---------

Copyright (c) 2020-2021 by the California Institute of Technology.  This code
is open-source software released under a 3-clause BSD license.  Please see the
file "LICENSE" for more information.
'''

from aenum import Enum, unique

@unique
class ExitCode(Enum):
    success        = 0, 'success -- program completed normally'
    no_net         = 1, 'no network detected'
    bad_arg        = 2, 'encountered a bad or missing value for an option'
    file_error     = 3, 'encountered a problem with a file'
    user_interrupt = 4, 'the user interrupted execution'
    exception      = 5, 'an exception or fatal error occurred'



'''
__main__: main command-line interface to eprints2archives

Authors
-------

Michael Hucka <mhucka@caltech.edu> -- Caltech Library

Copyright
---------

Copyright (c) 2020-2021 by the California Institute of Technology.  This code
is open-source software released under a 3-clause BSD license.  Please see the
file "LICENSE" for more information.
'''

from   bun import UI, inform, alert, alert_fatal, warn
from   commonpy.data_utils import timestamp
from   commonpy.file_utils import readable
from   commonpy.interrupt import config_interrupt, interrupt
import os
from   os import cpu_count
import plac
import sys

if __debug__:
    from sidetrack import set_debug, log

import eprints2archives
from   eprints2archives import print_version
from   .auth import AuthHandler
from   .exceptions import *
from   .exit_codes import ExitCode
from   .main_body import MainBody
from   .run_manager import RunManager
from   .services import service_names


# Main program.
# ......................................................................

@plac.annotations(
    api_url    = ('the URL for the REST API of the EPrints server',        'option', 'a'),
    dest       = ('send to destination service "D" (default: "all")',      'option', 'd'),
    error_out  = ('stop if encounter missing records or similar problems', 'flag',   'e'),
    force      = ('ask services to archive records even if already there', 'flag',   'f'),
    id_list    = ('list of EPrint record identifiers (can be a file)',     'option', 'i'),
    lastmod    = ('only get EPrints records modified after given date',    'option', 'l'),
    quiet      = ('do not print informational messages while working',     'flag',   'q'),
    user       = ('EPrints server user login name "U"',                    'option', 'u'),
    password   = ('EPrints server user password "P"',                      'option', 'p'),
    report     = ('save a report to file "R"',                             'option', 'r'),
    status     = ('only get records whose status is in the list "S"',      'option', 's'),
    threads    = ('number of threads to use (default: #cores/2)',          'option', 't'),
    no_color   = ('do not color-code terminal output',                     'flag',   'C'),
    no_keyring = ('do not store credentials in a keyring service',         'flag',   'K'),
    services   = ('print list of known archiving services and exit',       'flag',   'S'),
    version    = ('print version info and exit',                           'flag',   'V'),
    debug      = ('write detailed trace to "OUT" ("-" means console)',     'option', '@'),
)

def main(api_url = 'A', dest = 'D', error_out = False, force = False,
         id_list = 'I', lastmod = 'L', quiet = False, user = 'U',
         password = 'P', report = 'R', status = 'S', threads = 'T',
         no_color = False, no_keyring = False,
         services = False, version = False, debug = 'OUT'):
    '''eprints2archives sends EPrints content to web archiving services.

This program contacts a given EPrints server, obtains the list of documents it
serves (optionally modified based on selectors such as date), determines the
URLs for the document pages on the EPrints server, and sends the URLs to
archiving sites such as the Internet Archive.  This helps preserve the
EPrints server contents in third-party archiving sites.

Specifying which EPrints server to contact
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This program contacts an EPrints REST server whose network API is accessible
at the URL given by the command-line option -a (or /a on Windows). A typical
EPrints server URL has the form "https://server.institution.edu/rest". This
program will automatically add "/eprint" to the URL path, so when writing the
URL after option -a, omit the trailing "/eprint" part of the URL.

Accessing some EPrints servers via the API requires supplying a user login and
password to the server. By default, this program uses the operating system's
keyring/keychain functionality to get a user name and password. If the
information does not exist from a previous run of eprints2archives, it will
ask the user interactively for the user name and password, and (unless the -K
or /K argument is given) store them in the user's keyring/keychain so that it
does not have to ask again in the future. It is also possible to supply the
information directly on the command line using the -u and -p options (or /u
and /p on Windows), but this is discouraged because it is insecure on
multiuser computer systems. (However, if you need to reset the user name and/or
password for some reason, use -u with a user name and let it prompt for a
password again.)

If a given EPrints server does not require a user name and password, supply
blank values when prompted for them by eprints2archives and do not use the
options -u or -p. (Empty user name and password are allowed values.)

Specifying which records to send
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

The EPrints records to be sent to the web archiving services will be limited to
the records indicated by the option -i (or /i on Windows). If no -i option is
given, this program will use all the records available at the given EPrints
server. The value of -i can be one or more integers separated by commas
(e.g., -i 54602,54604), or a range of numbers separated by a dash (e.g.,
-i 1-100, which is interpreted as the list of numbers 1, 2, ..., 100 inclusive),
or some combination thereof. The value of the option -i can also be a file, in
which case, the file is read to get a list of identifiers.

If the -l option (or /l on Windows) is given, the records will be additionally
filtered to return only those whose last-modified date/time stamp is no older
than the given date/time description. Valid descriptors are those accepted
by the Python dateparser library. Make sure to enclose descriptions within
single or double quotes. Examples:

  eprints2archives -l "2 weeks ago" -a ....
  eprints2archives -l "2014-08-29"  -a ....
  eprints2archives -l "12 Dec 2014" -a ....
  eprints2archives -l "July 4, 2013" -a ....

If the -s option (or /s on Windows) is given, the records will also be filtered
to include only those whose eprint_status element value is one of the listed
status codes. Comparisons are done in a case-insensitive manner. Putting a
caret character ("^") in front of the status (or status list) negates the
sense, so that eprints2archives will only keep those records whose eprint_status
value is *not* among those given. Examples:

  eprints2archives -s archive -a ...
  eprints2archives -s ^inbox,buffer,deletion -a ...

Both lastmod and status filering are done after the -i argument is processed.

By default, if an error occurs when requesting a record from the EPrints
server, eprints2archives will keep going and not stop execution. Common causes
of errors include missing records implied by the arguments to -i, missing files
associated with a given record, and files inaccessible due to permissions
errors. If the option -e (or /e on Windows) is given, eprints2archives will
instead stop upon encountering a missing record, or missing file within a
record, or similar errors. The default is to merely issue warnings when missing
records are encountered because this is less frustrating for most use-cases.

How URLs are constructed
~~~~~~~~~~~~~~~~~~~~~~~~

eprints2archives always tries to construct 3 URLs for every EPrint record and
verifies that they exist on the EPrints server. The URLs are as follows, where
N is the id number of the EPrint record:

  1. https://SERVER/N
  2. https://SERVER/id/eprint/N
  3. The value of the field official_url (if any) in the EPrint record.

The first two typically go to the same page on an EPrint server, but web
archiving services have no direct mechanism to indicate that a given URL is an
alias or redirection for another, so they need to be sent as separate URLs. On
the other hand, the value of official_url may be an entirely different URL,
which may or may not go to the same location as one of the others.

Next, if no selection or filtering is applied (i.e., none of the options -i,
-l or -s are given to eprints2archives), eprints2archives gathers additional
URLs as follows (where SERVER is the server hostname + post number, if any):

  * https://SERVER
  * https://SERVER/view
  * The set of pages https://SERVER/view/X, where each X is obtained by
    parsing the HTML of https://SERVER/view and extracting links to pages
    under https://SERVER/view/
  * The set of pages https://SERVER/view/X/Y, where each Y is obtained by
    parsing the HTML of https://SERVER/view/X and extracting links to pages
    under https://SERVER/view/X

On the other hand, if selection and/or filtering _are_ in effect (i.e., if any
of the options -i, -l and/or -s are used), then eprints2archives ONLY
extracts the URLs https://SERVER/view/X/N.html for every EPrints identifier N
selected or left after filtering (if such URLs exist on the server).

Specifying where to send records
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Eprints2archives has a set of built-in adapters to interact with a number of
known public web archiving services. To learn which services eprints2archives
knows about, use the option -S (or /S on Windows). By default, the program will
send EPrints record URLs to all the known services. The option -d (or /d on
Windows) can be used to select one or a list of destination services instead.
Lists of services should be separated by commas with no spaces between them;
e.g., "internetarchive,archivetoday".

By default, eprints2archives will only ask a service to archive a copy of an
EPrints record if the service does not already have an archived copy.  This
makes sense because EPrints records usually change infrequently, and there's
little point in repeatedly asking web archives to make new archives.  However,
if you have reason to want the web archives to re-archive EPrints records, you
can use the option -f (or /f on Windows).

Eprints2archives will use parallel process threads to query the EPrints server
as well as to send records to archiving services.  By default, the maximum
number of threads used is equal to 1/2 of the number of cores on the computer
it is running on. The option -t (or /t on Windows) can be used to change this
number.  Eprints2archives will always use only one thread per web archiving
service (and since there are only a few services, only a few threads are usable
during that phase of operation), but a high number of threads can be helpful
to speed up the initial data gathering step from the EPrints server.

To save a report of the articles sent to archiving services, you can use the
option -r (/r on Windows) followed by a file name.

Note that there is a lag between when web archives such as Internet Archive
receive a URL submission and when a saved copy is made available from the
archive.  (In the case of Internet Archive, it is 3-10 hours.)  If you cannot
find a given EPrints page in an archive shortly after running eprints2archives,
it may be because not enough time has passed -- try again later.

Return values
~~~~~~~~~~~~~

This program exits with a return code of 0 if no problems are encountered.
It returns a nonzero value otherwise, following the conventions used in shells
such as bash which only understand return code values of 0 to 255. The
following table lists the possible return values:

    0 = success -- program completed normally
    1 = no network detected -- cannot proceed
    2 = encountered a bad or missing value for an option
    3 = file error -- encountered a problem with a file
    4 = the user interrupted the program's execution
    5 = an exception or fatal error occurred

Other command-line arguments
~~~~~~~~~~~~~~~~~~~~~~~~~~~~

eprints2archives will print messages as it works. To limit the messages
to warnings and errors only, use the option -q (or /q on Windows). Also,
output is color-coded by default unless the -C option (or /C on Windows) is
given; this option can be helpful if the color control signals create
problems for your terminal emulator.

If given the -@ argument (/@ on Windows), this program will output a detailed
trace of what it is doing, and will also drop into a debugger upon the
occurrence of any errors. The debug trace will be written to the given
destination, which can be a dash character (-) to indicate console output, or
a file path.

If given the -V option (/V on Windows), this program will print the version
and other information, and exit without doing anything else.

Command-line options summary
~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'''
    # Preprocess arguments and handle early exits -----------------------------

    debugging = debug != 'OUT'
    if debugging:
        if __debug__: set_debug(True, debug, extra = '%(threadName)s')
        from rich.traceback import install as install_rich_traceback
        install_rich_traceback()
        import faulthandler
        faulthandler.enable()

    if version:
        print_version()
        exit(0)

    if services:
        print('Known services:', ', '.join(service_names()))
        exit(0)

    # Do the real work --------------------------------------------------------

    if __debug__: log('='*8 + f' started {timestamp()} ' + '='*8)
    ui = manager = exception = None
    try:
        ui = UI('eprints2archives', 'send EPrints records to web archives',
                use_color = not no_color, be_quiet = quiet)
        auth = AuthHandler(prompter = ui.login_details,
                           user = None if user == 'U' else user,
                           pswd = None if password == 'P' else password,
                           use_keyring = not no_keyring)
        body = MainBody(api_url = None if api_url == 'A' else api_url,
                        id_list = None if id_list == 'I' else id_list,
                        lastmod = None if lastmod == 'L' else lastmod,
                        status  = 'any' if status == 'S' else status,
                        dest    = 'all' if dest == 'D' else dest,
                        threads = max(1, cpu_count()//2 if threads == 'T' else int(threads)),
                        auth_handler = auth,
                        quit_on_error = error_out,
                        force = force,
                        report_file = None if report == 'R' else report)
        config_interrupt(body.stop, UserCancelled(ExitCode.user_interrupt))
        manager = RunManager()
        manager.run(ui, body)
        exception = body.exception
    except (KeyboardInterrupt, UserCancelled) as ex:
        # In Python, the main thread (i.e., this one) is the one that gets ^C.
        alert('Quit received; shutting down ...')
        interrupt()
        manager.stop()
        exception = sys.exc_info()
    except Exception as ex:
        # MainBody exceptions are caught in its thread, so this is something else.
        exception = sys.exc_info()

    # Try to deal with exceptions gracefully ----------------------------------

    exit_code = ExitCode.success
    if exception:
        if type(exception[1]) == CannotProceed:
            exit_code = exception[1].args[0]
        elif type(exception[1]) in [KeyboardInterrupt, UserCancelled]:
            if __debug__: log(f'received {exception[1].__class__.__name__}')
            exit_code = ExitCode.user_interrupt
        else:
            exit_code = ExitCode.exception
            from traceback import format_exception
            msg = str(exception[1])
            details = ''.join(format_exception(*exception))
            if __debug__: log(f'Exception: {msg}\n{details}')
            if debugging:
                import pdb; pdb.set_trace()
            if manager:
                manager.stop()
    if __debug__: log('_'*8 + f' stopped {timestamp()} ' + '_'*8)
    if exit_code == ExitCode.user_interrupt:
        # This is a sledgehammer, but it kills everything, including ongoing
        # network get/post. I have not found a more reliable way to interrupt.
        os._exit(exit_code.value[0])
    else:
        exit(exit_code.value[0])


# Main entry point.
# .............................................................................

# On windows, we want plac to use slash intead of hyphen for cmd-line options.
if sys.platform.startswith('win'):
    main.prefix_chars = '/'

# The following entry point definition is for the console_scripts keyword
# option to setuptools.  The entry point for console_scripts has to be a
# function that takes zero arguments.
def console_scripts_main():
    plac.call(main)

# The following allows users to invoke this using "python3 -m eprints2archives".
if __name__ == '__main__':
    plac.call(main)


# For Emacs users
# ......................................................................
# Local Variables:
# mode: python
# python-indent-offset: 4
# End:



'''
upload_status.py: notification of the current status of an upload task

Authors
-------

Michael Hucka <mhucka@caltech.edu> -- Caltech Library

Copyright
---------

Copyright (c) 2020-2021 by the California Institute of Technology.  This code is
open-source software released under a 3-clause BSD license.  Please see the
file "LICENSE" for more information.
'''

from enum import Enum

class ServiceStatus(Enum):
    RUNNING           = 1               # Currently uploading.
    PAUSED_RATE_LIMIT = 2               # Paused due to hitting a rate limit.
    PAUSED_ERROR      = 3               # Paused due to an error.
    UNAVAILABLE       = 4               # Server doesn't want to let us use it.



'''
exceptions.py: exceptions defined by eprints2archives

Authors
-------

Michael Hucka <mhucka@caltech.edu> -- Caltech Library

Copyright
---------

Copyright (c) 2020-2021 by the California Institute of Technology.  This code
is open-source software released under a 3-clause BSD license.  Please see the
file "LICENSE" for more information.
'''


# Base class.
# .............................................................................
# The base class makes it possible to use a single test to distinguish between
# exceptions generated by eprints2archives code and exceptions generated by
# something else.

class Eprints2ArchivesException(Exception):
    '''Base class for eprints2archives exceptions.'''
    pass


# Exception classes.
# .............................................................................

class CannotProceed(Eprints2ArchivesException):
    '''A recognizable condition caused an early exit from the program.'''
    pass

class UserCancelled(Eprints2ArchivesException):
    '''The user elected to cancel/quit the program.'''
    pass

class CorruptedContent(Eprints2ArchivesException):
    '''Content corruption has been detected.'''
    pass

class ServerError(Eprints2ArchivesException):
    '''Unrecoverable problem involving the EPrints server.'''
    pass

class InternalError(Eprints2ArchivesException):
    '''Unrecoverable problem involving eprints2archives itself.'''
    pass



'''
internetarchive.py: interface to the Wayback Machine at the Internet Archive.

Authors
-------

Michael Hucka <mhucka@caltech.edu> -- Caltech Library

Copyright
---------

Copyright (c) 2020-2021 by the California Institute of Technology.  This code
is open-source software released under a 3-clause BSD license.  Please see the
file "LICENSE" for more information.
'''

from   bun import inform, alert, alert_fatal, warn
from   commonpy.exceptions import NoContent, ServiceFailure, RateLimitExceeded
from   commonpy.interrupt import interrupted, wait
from   commonpy.network_utils import net, hostname
from   humanize import intcomma
import requests
from   time import sleep

if __debug__:
    from sidetrack import log

from ..exceptions import *

from .base import Service
from .timemap import timemap_as_dict, timemap_mementos
from .upload_status import ServiceStatus


# Constants.
# .............................................................................

_MAX_RETRIES = 8
'''Maximum number of times we retry before we give up.'''

_RETRY_SLEEP = 60
'''Time in seconds we pause if we get repeated errors.  This pause is on top
of the maximum time reached after exponential back-off by our underlying
network code, and is used to pause before the whole scheme is retried again.
This needs to be a long time because if we invoke our own retry loop, it means
we've already been trying for a considerable amount of time and the root cause
may be significant.'''

_RATE_LIMIT_SLEEP = 10
'''Time in seconds we pause if we hit the rate limit.  This is handled
separately from error conditions.'''


# Classes.
# .............................................................................

class InternetArchive(Service):
    label = 'internetarchive'
    name  = 'Internet Archive'
    color = 'white'

    # Public methods ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    def save(self, url, notify, force = False):
        '''Ask the service to save "url".'''
        if force:
            # If we're forcing a send, we don't care how many there exist.
            added = self._archive(url, notify)
            return (added, -1)

        timemap = self._timemap_for_url(url, notify)
        if timemap:
            mementos = timemap_mementos(timemap)
            if __debug__: log(f'{self.name} returned {len(mementos)} mementos for {url}')
            return (False, len(mementos))
        else:
            if __debug__: log(f'{self.name} returned no mementos for {url}')
            added = self._archive(url, notify)
            return (added, 0)


    # Internal methods ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    def _uniform(self, url):
        return str(url).strip().replace(' ', '_')


    def _timemap_for_url(self, url, notify):
        '''Returns a timemap, in the form of a dict.'''
        if __debug__: log(f'asking {self.name} for info about {url}')

        action_url = 'https://web.archive.org/web/timemap/link/' + self._uniform(url)
        (response, error) = net('get', action_url, handle_rate = False)
        if not error and response:
            if __debug__: log('converting TimeMap to dict')
            return timemap_as_dict(response.text, skip_errors = True)
        elif isinstance(error, NoContent):
            if __debug__: log(f'no content for {url}')
            return {}
        elif isinstance(error, RateLimitExceeded):
            if __debug__: log(f'{self.name} rate limit; pausing {_RATE_LIMIT_SLEEP}s')
            notify(ServiceStatus.PAUSED_RATE_LIMIT)
            wait(_RATE_LIMIT_SLEEP)
            notify(ServiceStatus.RUNNING)
            return self._timemap_for_url(url, notify)
        elif isinstance(error, ServiceFailure):
            # Our underlying network code will retry most of these cases, so if
            # we get here, the timemap request is being rejected for some reason.
            if __debug__: log(f'no timemap due to "{error}"')
            return {}
        else:
            if __debug__: log(f'got "{error}"')
            raise error


    def _archive(self, url, notify, retry = 0):
        if __debug__: log(f'asking {self.name} to save {url}')
        if retry > 0:
            if __debug__: log(f'this is retry #{retry}')
        payload = {'url': url, 'capture_all': 'on'}
        action_url = 'https://web.archive.org/save/' + self._uniform(url)
        (response, error) = net('post', action_url, handle_rate = False, data = payload)
        if not error:
            if __debug__: log(f'save request accepted by {self.name} for {url}')
            return True
        elif isinstance(error, RateLimitExceeded):
            if __debug__: log(f'{self.name} rate limit; pausing {_RATE_LIMIT_SLEEP}s')
            notify(ServiceStatus.PAUSED_RATE_LIMIT)
            wait(_RATE_LIMIT_SLEEP)
            notify(ServiceStatus.RUNNING)
            if __debug__: log(f'trying again recursively for {url}')
            return self._archive(url, notify)
        elif isinstance(error, ServiceFailure):
            # Our underlying network code will retry most of these cases, so if
            # we get here, the save request is being rejected for some reason.
            if __debug__: log(f'save request rejected by {self.name} for {url}')
            return False
        else:
            if __debug__: log(f'save request resulted in an error: {str(error)}')
            # Our underlying net(...) function will retry automatically in
            # the face of problems, but will give up eventually.  Sometimes
            # IA errors are temporary, so we pause for even longer & retry.
            retry += 1
            retries_left = _MAX_RETRIES - retry
            if __debug__: log(f'we have {retries_left} retries left')
            if retry == 1:
                # Might have been a transient server-unavailable type of error.
                if __debug__: log(f'retrying once without pause')
                return self._archive(url, notify, retry)
            elif retries_left > 0:
                # Subtract 1 b/c we try without pause once, before we land here.
                if __debug__: log(f'pausing due to multiple retries')
                sleeptime = _RETRY_SLEEP * pow(retry - 1, 2)
                notify(ServiceStatus.PAUSED_ERROR)
                wait(sleeptime)
                notify(ServiceStatus.RUNNING)
                return self._archive(url, notify, retry)
            else:
                if __debug__: log(f'retry limit reached for {self.name}.')
                raise error
        return False



'''
timemap.py: parse TimeMaps

This file was originally timemap.py from the Off-Topic Memento Toolkit,
https://github.com/oduwsdl/off-topic-memento-toolkit/blob/master/otmt/timemap.py
retrieved from GitHub on 2020-07-29.  The git commit hash at the time I copied
the file was e714c8ad3c41221b83b24fda1979844d03d83726.

The license for OTMT at the time timemap.py was copied was the MIT license,
https://github.com/oduwsdl/off-topic-memento-toolkit/blob/master/LICENSE.txt

Since copying the original, I've made some alterations to this source code.
The git history can be consulted to find out what the changes have been.

Authors (original)
------------------

Shawn M. Jones <jones.shawn.m@gmail.com> -- Old Dominion University

Authors (subsequent modifications)
----------------------------------

Michael Hucka <mhucka@caltech.edu> -- Caltech Library
'''

from datetime import datetime

from ..exceptions import CorruptedContent


def timemap_as_dict(timemap_text, skip_errors = False):
    '''A function to convert the link format TimeMap text into a Python
    dictionary that closely resembles the JSON specified at:
    http://mementoweb.org/guide/timemap-json/

    There is one difference: the value of the datetime attribute is
    an actual Python datetime object.

    One can set skip_errors to True in order to skip errors in processing the
    TimeMap, but use with caution as it can lead to unpredictable behavior.
    '''

    def process_local_dict(local_dict, working_dict):
        first = False
        last = False

        for uri in local_dict:

            if "rel" not in local_dict[uri]:
                raise ValueError("Missing 'rel' element in timemap")

            relation = local_dict[uri]["rel"]

            if relation == "original":
                working_dict["original_uri"] = uri

            elif relation == "timegate":
                working_dict["timegate_uri"] = uri

            elif relation == "self":
                working_dict["timemap_uri"] = {}
                working_dict["timemap_uri"]["link_format"] = uri

            elif "memento" in relation:
                working_dict.setdefault("mementos", {})

                if "first" in relation:
                    working_dict["mementos"]["first"] = {}
                    working_dict["mementos"]["first"]["uri"] = uri
                    first = True

                if "last" in relation:
                    working_dict["mementos"]["last"] = {}
                    working_dict["mementos"]["last"]["uri"] = uri
                    last = True

                working_dict["mementos"].setdefault("list", [])

                local_memento_dict = {
                    "datetime": None,
                    "uri": uri
                }

            if "datetime" in local_dict[uri]:

                mdt = datetime.strptime(local_dict[uri]["datetime"],
                    "%a, %d %b %Y %H:%M:%S GMT")

                local_memento_dict["datetime"] = mdt

                working_dict["mementos"]["list"].append(local_memento_dict)

                if first:
                    working_dict["mementos"]["first"]["datetime"] = mdt

                if last:
                    working_dict["mementos"]["last"]["datetime"] = mdt

        return working_dict

    dict_timemap = {}

    # current_char = ""
    uri = ""
    key = ""
    value = ""
    local_dict = {}
    state = 0
    charcount = 0

    for character in timemap_text:
        charcount += 1

        if state == 0:

            local_dict = {}
            uri = ""

            if character == '<':
                state = 1
            elif character.isspace():
                pass
            else:
                if not skip_errors:
                    raise CorruptedContent(
                        "issue at character {} while looking for next URI"
                        .format(charcount))

        elif state == 1:

            if character == '>':
                # URI should be saved by this point
                state = 2
                uri = uri.strip()
                local_dict[uri] = {}
            else:
                uri += character

        elif state == 2:

            if character == ';':
                state = 3

            elif character.isspace():
                pass

            else:
                if not skip_errors:
                    raise CorruptedContent(
                        "issue at character {} while looking for relation"
                        .format(charcount))

        elif state == 3:

            if character == '=':
                state = 4
            else:
                key += character

        elif state == 4:

            if character == ';':
                state = 3
            elif character == ',':
                state = 0
                process_local_dict(local_dict, dict_timemap)
            elif character == '"':
                state = 5
            elif character.isspace():
                pass
            else:
                if not skip_errors:
                    raise CorruptedContent(
                        "issue at character {} while looking for value"
                        .format(charcount))

        elif state == 5:

            if character == '"':
                state = 4

                key = key.strip()
                value = value.strip()
                local_dict[uri][key] = value
                key = ""
                value = ""

            else:
                value += character

        else:

            if not skip_errors:
                raise CorruptedContent(
                    "discovered unknown state while processing TimeMap")

    try:
        process_local_dict(local_dict, dict_timemap)
    except Exception as ex:
        if not skip_errors:
            raise CorruptedContent(
                "unexpected construction in timemap: " + str(ex))
        else:
            return dict()

    return dict_timemap


def timemap_mementos(timemap_dict):
    '''Take a timemap as dict and return the list of mementos therein.'''
    if not isinstance(timemap_dict, dict):
        raise ValueError(f'Expected a dict but got {type(timemap_dict)}.')
    if 'mementos' in timemap_dict and 'list' in timemap_dict['mementos']:
        return timemap_dict['mementos']['list']
    else:
        return []



'''
archivetoday.py: interface to Archive.Today.

Much of the basic algorithm for interacting with Archive.Today was initially
taken from ArchiveNow (https://github.com/oduwsdl/archivenow), from the
Web Science and Digital Libraries Research Group at Old Dominion University.
The ArchiveNow source file used was the following, and it had commit hash
b4e88b2 at the time of this writing (2020-08-06):

https://github.com/oduwsdl/archivenow/blob/master/archivenow/handlers/is_handler.py

The authors & contributors of that specific code were (again at the time of
this writing) Mohamed Aturban, Shawn M. Jones, "veloute", and "evil-wayback".
The license for ArchiveNow at the time timemap.py was copied was the MIT license,
https://github.com/oduwsdl/archivenow/blob/master/LICENSE

Authors
-------

Michael Hucka <mhucka@caltech.edu> -- Caltech Library

Copyright
---------

Copyright (c) 2020-2021 by the California Institute of Technology.  This code
is open-source software released under a 3-clause BSD license.  Please see the
file "LICENSE" for more information.
'''

from   bun import inform, alert, alert_fatal, warn
from   collections import OrderedDict
from   commonpy.exceptions import NoContent, ServiceFailure
from   commonpy.interrupt import interrupted, wait
from   commonpy.network_utils import net, hostname
from   humanize import intcomma
import requests
from   time import sleep
import urllib
from   urllib.parse import quote_plus, urlencode

if __debug__:
    from sidetrack import log

from ..exceptions import *

from .base import Service
from .timemap import timemap_as_dict, timemap_mementos
from .upload_status import ServiceStatus


# Constants.
# .............................................................................

# Archive.today (= Archive.is and other names) deliberately uses multiple
# domains.  The author explained the situation in a blog posting of 2020-08-01
# (https://blog.archive.today/post/625241784047550464):
#
# """""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
# No single domain is reliable and I have no means to enforce control on
# each domain.
#
# * archive.today - threatened with confiscation
# http://blog.archive.today/post/116913927371, also a troll attack caused
# service interruption https://blog.archive.today/post/138982909006
#
# * archive.is - threatened with confiscation
# https://twitter.com/archiveis/status/1081276424781287427, asked not to use
# “archive.IS” for branding (that’s why you see “archive.TODAY” in the top-left
# corner; although many people remembered it as “archive.IS” and refer it so)
#
# * archive.fo - threatened with confiscation
# https://twitter.com/archiveis/status/1188222460598116353
#
# * archive.li - attacked by trolls impersonating police, caused few days
# service interruption https://twitter.com/archiveis/status/956025540028268547
#
# * archive.ec - attacked by trolls causing service interruption and finally
# lost https://twitter.com/archiveis/status/1093608363647291393
#
# * archive.vn - ok so far
#
# * archive.ph - ok so far
#
# * archive.md - ok so far
#
# * a nice domain unrelated to archive - one day whois started showing
# someone’s else information and the registrar did not response, the
# domain was lost
#
# """""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
#
# A list of the URLs is available from the Wikipedia page for Archive.Today:
# https://en.wikipedia.org/wiki/Archive.today

_HOSTS = ['archive.li', 'archive.vn','archive.fo', 'archive.md', 'archive.ph',
          'archive.today', 'archive.is']
'''Alternative hosts by which Archive.Today is known.'''

_USER_AGENT = "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.116 Safari/537.36"
'''User agent we pretend to be when contacting Archive.Today.'''

_MAX_RETRIES = 8
'''Maximum number of times we retry before we give up.'''

_RETRY_SLEEP = 60
'''Time in seconds we pause if we get repeated errors.  This pause is on top
of the maximum time reached after exponential back-off by our underlying
network code, and is used to pause before the whole scheme is retried again.
This needs to be a long time because if we invoke our own retry loop, it means
we've already been trying for a considerable amount of time and the root cause
may be significant.'''

_RATE_LIMIT_SLEEP = 300
'''Time in seconds we pause if we hit the rate limit.  This is handled
separately from error conditions.'''


# Classes.
# .............................................................................

class ArchiveToday(Service):
    label = 'archive.today'
    name = 'Archive.today'
    color = 'deep_sky_blue1'

    _host = None                        # archive.il or archive.is or ...
    _sid  = None                        # current submit id value
    _available = True                   # did we manage to find a host?

    # Public methods ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    def save(self, url, notify, force = False):
        '''Ask the service to save "url".'''
        if self._available and self._host is None:
            self._host = self._archive_host()
            self._available = self._host is not None
        if not self._available:
            notify(ServiceStatus.UNAVAILABLE)
            return (False, -1)

        if force:
            # If we're forcing a send, we don't care how many copies exist.
            added = self._archive(url, notify)
            return (added, -1)

        timemap = self._timemap_for_url(url, notify)
        if timemap:
            mementos = timemap_mementos(timemap)
            if __debug__: log(f'there are {len(mementos)} mementos for {url}')
            return (False, len(mementos))
        else:
            if __debug__: log(f'{self.name} returned no mementos for {url}')
            added = self._archive(url, notify)
            return (added, 0)


    # Internal methods ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    def _uniform(self, url):
        return str(url).strip().replace(' ', '_')


    def _timemap_for_url(self, url, notify):
        '''Returns a timemap, in the form of a dict.'''

        headers = {"User-Agent": _USER_AGENT}
        action_url = f'https://{self._host}/timemap/' + self._uniform(url)
        (response, error) = net('get', action_url, headers = headers)
        if not error and response:
            if __debug__: log('converting TimeMap to dict')
            return timemap_as_dict(response.text, skip_errors = True)
        elif isinstance(error, NoContent):
            return {}
        elif isinstance(error, ServiceFailure) and response:
            # Archive.today doesn't return code 429 when you hit the rate limit
            # and instead throws code 503.  See author's posting of 2020-08-04:
            # https://blog.archive.today/post/625519838592417792
            if response.status_code == 503:
                if __debug__: log(f'{self.name} rate limit; pausing {_RATE_LIMIT_SLEEP}s')
                notify(ServiceStatus.PAUSED_RATE_LIMIT)
                wait(_RATE_LIMIT_SLEEP)
                notify(ServiceStatus.RUNNING)
                return self._timemap_for_url(url, notify)
        else:
            raise error


    def _archive_host(self):
        headers = {"User-Agent": _USER_AGENT}
        if __debug__: log(f'looking for active {self.name} host')
        archive_host = None
        for host in _HOSTS:
            # headers['host'] = host
            test_url = f'https://{host}/'
            (response, error) = net('get', test_url, headers = headers)
            if not error:
                if __debug__: log(f'Archive.Today host is currently {host}')
                archive_host = host
                break
            elif isinstance(error, ServiceFailure) and response.status_code == 503:
                continue
            else:
                raise error
        if archive_host is None:
            return None
        try:
            html = str(response.content)
            # Gnarly line of code from ArchiveNow.
            self._sid = html.split('name="submitid', 1)[1].split('value="', 1)[1].split('"', 1)[0]
        except:
            raise InternalError(f'Unable to parse {self.name} page')
        return archive_host


    def _archive(self, url, notify, retry = 0):
        # Basic idea and algorithm taken from ArchiveNow.  We iterate over the
        # various domain names that Archive.today uses, because some of them
        # stop responding and others start responding, and we never know which
        # one will work.  If you try to use /submit on archive.today itself,
        # you get an error about too many redirects.  (Not sure if that's
        # deliberate on their part or just a side-effect of what they're doing
        # with their hosts.
        if __debug__: log(f'will ask {self.name} to save {url}')

        action_url = f'https://{self._host}/submit/'
        headers = {"User-Agent": _USER_AGENT}
        headers['host'] = self._host
        # The order of the content of the post body matters to Archive.today.
        payload = OrderedDict({'submitid': self._sid, 'url': url})
        (response, error) = net('post', action_url, handle_rate = False,
                                headers = headers, data = payload)

        if not error:
            if 'Refresh' in response.headers:
                try:
                    saved_url = str(response.headers['Refresh']).split(';url=')[1]
                    if __debug__: log(f'{self.name} saved URL as {saved_url}')
                    return True
                except:
                    raise InternalError('Unexpected response from {self.name}')
            elif 'Location' in response.headers:
                saved_url = response.headers['Location']
                if __debug__: log(f'{self.name} saved URL as {saved_url}')
                return True
            else:
                for h in response.history:
                    if 'Location' in h.headers:
                        saved_url = h.headers['Location']
                        if __debug__: log(f'{self.name} saved URL as {saved_url}')
                        return True
            raise InternalError(f'{self.name} returned unexpected response')
        else:
            # Archive.today doesn't return code 429 when you hit the rate limit
            # and instead throws code 503.  See author's posting of 2020-08-04:
            # https://blog.archive.today/post/625519838592417792
            if isinstance(error, ServiceFailure):
                if __debug__: log(f'{self.name} rate limit; pausing {_RATE_LIMIT_SLEEP}s')
                notify(ServiceStatus.PAUSED_RATE_LIMIT)
                wait(_RATE_LIMIT_SLEEP)
                notify(ServiceStatus.RUNNING)
                return self._archive(url, notify)

            # Our underlying net(...) function will retry automatically for
            # some recognizable temporary problems.  Others, we handle here.
            if __debug__: log(f'save request resulted in an error: {str(error)}')
            retry += 1
            retries_left = _MAX_RETRIES - retry
            if __debug__: log(f'we have {retries_left} retries left')
            if retries_left > 0:
                sleeptime = _RETRY_SLEEP * pow(retry, 2)
                warn(f'Got error from {self.name}; pausing for {intcomma(sleeptime)}s.')
                notify(ServiceStatus.PAUSED_ERROR)
                wait(sleeptime)
                notify(ServiceStatus.RUNNING)
                return self._archive(url, notify, retry)
            else:
                if __debug__: log(f'retry limit reached for {self.name}.')
                raise error
        return False



'''
run_manager.py: manage thread execution

Authors
-------

Michael Hucka <mhucka@caltech.edu> -- Caltech Library

Copyright
---------

Copyright (c) 2018-2021 by the California Institute of Technology.  This code
is open-source software released under a 3-clause BSD license.  Please see the
file "LICENSE" for more information.
'''

from   commonpy.interrupt import interrupt
from   pubsub import pub
import sys
from   threading import Thread

if __debug__:
    from sidetrack import log

from .exceptions import *


# Class definitions.
# .............................................................................
# This class abstracts away the details of starting both a worker thread and
# the user interface.  The latter typically involves starting a GUI
# interface, although those details are abstracted out to a separate UI class.
#
# The basic problem that this faces is that on macOS, the wx main loop cannot
# be started in a subthread: it must be started in the main thread.  However,
# the wx MainLoop() call is a blocking call, so as soon as we invoke it
# (whether directly or via the ui.start() call as done below), nothing further
# happens in the current thread until the wx app thread exits.  This prevents
# us from doing what would make the most sense, which is to use a subthread
# to start the wx MainLoop() so that we don't block, and can instead wait for
# the worker thread to finish.

class RunManager():
    '''Manager for overall program execution.'''

    def __init__(self):
        # Listen to stop messages.
        pub.subscribe(self.stop, "stop")


    def run(self, ui, worker):
        self._ui = ui
        self._worker = worker

        # On Windows, in Python 3.6+, ^C in a terminal window does not stop
        # execution (at least in my environment).  The following function
        # creates a closure with the worker object so that stop() can be called.

        if sys.platform == "win32":
            if __debug__: log('installing ctrl_handler for Windows')

            # This is defined here because we need the value of worker.ident.
            def ctrl_handler(event, *args):
                if __debug__: log('Keyboard interrupt received')
                from stopit import async_raise
                interrupt()
                async_raise(worker.ident, UserCancelled)
                worker.stop()

            import win32api
            win32api.SetConsoleCtrlHandler(ctrl_handler, True)

        # The worker must be a daemon thread when using the GUI; for CLI, it
        # must not be a daemon thread or else the program exits immediately.
        worker.daemon = ui.is_gui()

        # Must start worker before ui, because in the GUI case, ui.start()
        # will block until the UI thread exits.
        if __debug__: log('starting worker')
        worker.start()
        if __debug__: log('starting main UI loop')
        ui.start()
        if not ui.is_gui():
            if __debug__: log('waiting on worker thread')
            worker.join()
        if __debug__: log('stopping worker')
        worker.stop()


    def stop(self):
        if self._worker:
            if __debug__: log('calling stop() on worker')
            self._worker.stop()
            if __debug__: log('calling stop() on UI')
        self._ui.stop()



'''
auth.py: code to deal with authenticating the user to the service

Authors
-------

Michael Hucka <mhucka@caltech.edu> -- Caltech Library

Copyright
---------

Copyright (c) 2018-2021 by the California Institute of Technology.  This code
is open-source software released under a 3-clause BSD license.  Please see the
file "LICENSE" for more information.
'''

import getpass
import keyring
import sys

if sys.platform.startswith('win'):
    import keyring.backends
    from keyring.backends.Windows import WinVaultKeyring
if sys.platform.startswith('darwin'):
    import keyring.backends
    from keyring.backends.OS_X import Keyring

if __debug__:
    from sidetrack import log, logf


# Global constants.
# .............................................................................

_KEYRING = f'org.caltechlibrary.{__package__}'
'''The name of the keyring used to store server access credentials, if any.'''

_EMPTY = ''
'''Character used to store an empty login name or password.'''


# Exported class.
# .............................................................................

class AuthHandler():
    '''Class to use the command line to ask the user for login & password.'''

    def __init__(self, prompter = None, user = None, pswd = None, use_keyring = True):
        '''Initializes this authentication handler object.

        The main purpose of AuthHandler is to provide the method named
        "credentials()", which callers can use to ask the user for a login name
        and password for a service.  In support of that goal, this init method
        takes arguments that set the user interface as well as default values
        for a user name and passsword, and whether to use a keyring.

        "prompter" must be a function that takes 3 arguments (a prompt string,
        an optional user name, and an optional password) and asks the user for
        login credentials, using the values of "user" and "pswd" as defaults.

        "user" can be a user name, if it is known at the time of creation of
        this AuthHandler object.  It will be used as the optional user name
        in calls to the "prompter".

        "password" can be a password, if it is known at the time of creation
        of this AuthHandler object.  It will be used as the optional password
        in calls to the "prompter".

        "use_keyring" should be a Boolean to indicate whether the keyring
        should be used.  If the value is False, then the calls to the function
        credentials(...) do not retrieve or store values in the keyring.
        '''
        self._prompter = prompter
        self._user = user
        self._pswd = pswd
        self._use_keyring = use_keyring


    # Explanation about the weird way this is done: the Python keyring module
    # only offers a single function for setting a value; ostensibly, this is
    # intended to store a password associated with an identifier (a user name),
    # and this identifier is expected to be obtained some other way, such as by
    # using the current user's computer login name.  This poses 2 problems for us:
    #
    #  1. The user may want to use a different user name for the remote service,
    #  so we can't assume the user's computer login name is the same.  We also
    #  don't want to ask for the remote user name every time we need the
    #  information, because that can end up presenting a dialog to the user every
    #  time, which quickly becomes unbearably annoying.  This means we can't use
    #  a user-generated identifer to access the keyring value -- we have to
    #  invent a value, and then store the user's name for the remote service as
    #  part of the value we store.
    #
    #  2. We need to store several pieces of information, not just a password,
    #  but the Python keyring module interface (and presumably most system
    #  keychains) does not allow anything but a string value.  The hackacious
    #  solution taken here is to concatenate several values into a single string
    #  used as the actual value stored.  The individual values are separated by a
    #  character that is unlikely to be part of any user-typed value.
    #
    # Explanation of values and their meanings:
    #   None                  => no value stored
    #   _EMPTY                => value stored and it is deliberately an empty string
    #   not None & not _EMPTY => an actual value

    def credentials(self, server, user = None, password = None):
        '''Returns a tuple of user, password, and a Boolean indicating
        whether the user cancelled the dialog.
        '''
        # If user name is provided and password is provided, use them.
        # If no user name is provided, look in keyring & prompt if not there.

        if __debug__: logf('keyring {}', 'enabled' if self._use_keyring else 'disabled')
        tmp_user = user if user is not None else self._user
        tmp_pswd = password if password is not None else self._pswd
        if (tmp_user is None and tmp_pswd is None) and self._use_keyring:
            # We weren't given a user name, but we can look in the keyring.
            if __debug__: log(f'getting login & password for {server} from keyring')
            (_, k_user, k_pswd) = keyring_credentials(server)
            tmp_user = '' if k_user == _EMPTY else k_user
            tmp_pswd = '' if k_pswd == _EMPTY else k_pswd
        cancel = False
        if tmp_user is None or tmp_pswd is None:
            # If user is not None but password is, we still have to prompt.
            prompt = f'User credentials for {server}'
            tmp_user, tmp_pswd, cancel = self._prompter(prompt, tmp_user, tmp_pswd)
        if cancel:
            return tmp_user, tmp_pswd, True
        if self._use_keyring:
            # Save the credentials.
            if __debug__: log('saving new credentials to keyring')
            save_keyring_credentials(server, tmp_user or _EMPTY, tmp_pswd or _EMPTY)
        self._user = tmp_user
        self._pswd = tmp_pswd
        return self._user, self._pswd, False


# Utility functions
# .............................................................................

def keyring_credentials(server):
    '''Looks up credentials stored for the given "server".'''
    if sys.platform.startswith('win'):
        keyring.set_keyring(WinVaultKeyring())
    if sys.platform.startswith('darwin'):
        keyring.set_keyring(Keyring())
    value = keyring.get_password(_KEYRING, server)
    return _decoded(value) if value else (None, None, None)


def save_keyring_credentials(server = '', user = '', pswd = ''):
    '''Saves the "user" and "password" for "server".'''
    if sys.platform.startswith('win'):
        keyring.set_keyring(WinVaultKeyring())
    if sys.platform.startswith('darwin'):
        keyring.set_keyring(Keyring())
    keyring.set_password(_KEYRING, server, _encoded(server, user, pswd))


_sep = ''
'''Character used to separate multiple actual values stored as a single
encoded value string.  This character is deliberately chosen to be something
very unlikely to be part of a legitimate string value typed by user at a
shell prompt, because control-c is normally used to interrupt programs.
'''


def _encoded(server, user, pswd):
    return f'{server}{_sep}{user}{_sep}{pswd}'


def _decoded(value_string):
    return tuple(value_string.split(_sep))



'''
main_body.py: main body logic for this application

Authors
-------

Michael Hucka <mhucka@caltech.edu> -- Caltech Library

Copyright
---------

Copyright (c) 2018-2021 by the California Institute of Technology.  This code
is open-source software released under a 3-clause BSD license.  Please see the
file "LICENSE" for more information.
'''

from   bun import inform, alert, alert_fatal, warn
from   commonpy.data_utils import DATE_FORMAT, slice, expanded_range, pluralized
from   commonpy.data_utils import timestamp, parsed_datetime
from   commonpy.exceptions import NoContent, AuthenticationFailure, ServiceFailure
from   commonpy.file_utils import writable
from   commonpy.interrupt import interrupted, raise_for_interrupts
from   commonpy.network_utils import network_available, netloc
from   concurrent.futures import ThreadPoolExecutor
from   humanize import intcomma
from   itertools import repeat
from   pydash import flatten
import re
from   rich.progress import Progress, BarColumn, TextColumn
import sys
from   threading import Thread
import time
from   validators.url import url as valid_url

if __debug__:
    from sidetrack import log

from .eprints import *
from .exceptions import *
from .exit_codes import ExitCode
from .services import ServiceStatus, service_names, service_interfaces, service_by_name


# Constants.
# .............................................................................

_PARALLEL_THRESHOLD = 2

_BAR = BarColumn(bar_width = None)
'''All our progress bars use the same kind of column.'''


# Class definitions.
# .............................................................................

class MainBody(Thread):
    '''Main body of eprints2archives implemented as a Python thread.'''

    def __init__(self, **kwargs):
        '''Initializes main body thread object but does not start the thread.'''
        Thread.__init__(self)

        # Assign parameters to self to make them available within this object.
        for key, value in kwargs.items():
            setattr(self, key, value)

        # We expose attribute "exception" that callers can use to find out if
        # the thread finished normally or with an exception.
        self.exception = None

        # Additional attributes we set later.
        self.user = None
        self.password = None

        # An unfortunate feature of Python's thread handling is that threads
        # don't get interrupt signals: if the user hits ^C, the parent thread
        # has to do something to interrupt the child threads deliberately.
        # We can't do that unless we keep a pointer to the executor and share
        # it between methods in this class.  Thus, the need for the following:
        self._executor = None
        self._futures = []


    def run(self):
        '''Run the main body.'''
        # In normal operation, this method returns after things are done and
        # leaves it to the user to exit the application via the control GUI.
        # If exceptions occur, we capture the stack context for the caller.
        if __debug__: log('running MainBody thread')

        try:
            self._do_preflight()
            self._do_main_work()
        except (KeyboardInterrupt, UserCancelled) as ex:
            if __debug__: log(f'got {type(ex).__name__}')
            self._report('Interrupted')
            self.exception = (ex, ex)
        except Exception as ex:
            if __debug__: log(f'exception in main body: {str(ex)}')
            self.exception = sys.exc_info()
            alert_fatal(f'Error occurred during execution:', details = str(ex))
        if __debug__: log('finished MainBody')


    def stop(self):
        '''Stop the main body thread.'''
        if self._executor:
            if __debug__: log('cancelling futures & shutting down executor')
            for f in self._futures:
                f.cancel()
            self._executor.shutdown(wait = False)
        else:
            if __debug__: log('no thread pool threads running')


    def _do_preflight(self):
        '''Check the option values given by the user, and do other prep.'''
        # We can't do anything without a network.
        if not network_available():
            alert_fatal('No network connection.')
            raise CannotProceed(ExitCode.no_net)

        hint = f'(Hint: use {"/" if sys.platform.startswith("wi") else "-"}h for help.)'

        # We can't do anything without the EPrints server URL.
        if self.api_url is None:
            alert_fatal(f'Must provide an EPrints API URL. {hint}')
            raise CannotProceed(ExitCode.bad_arg)

        if self.lastmod:
            try:
                self.lastmod = parsed_datetime(self.lastmod)
                self.lastmod_str = self.lastmod.strftime(DATE_FORMAT)
                if __debug__: log(f'parsed lastmod as {self.lastmod_str}')
            except Exception as ex:
                alert_fatal(f'Unable to parse lastmod value: "{str(ex)}". {hint}')
                raise CannotProceed(ExitCode.bad_arg)

        # It's easier to use None as an indication of no restriction (= 'any').
        if self.status:
            self.status = None if self.status == 'any' else re.split('(\W)', self.status)

        if self.dest == 'all':
            self.dest = service_interfaces()
        else:
            destination_list = self.dest.split(',')
            self.dest = [service_by_name(d) for d in destination_list]
            if None in self.dest:
                bad_dest = destination_list[self.dest.index(None)]
                alert_fatal(f'Unknown destination service "{bad_dest}". {hint}')
                raise CannotProceed(ExitCode.bad_arg)

        host = netloc(self.api_url)
        self.user, self.password, cancel = self.auth_handler.credentials(host)
        if cancel:
            raise UserCancelled
        if self.user:
            inform(f'Using login credentials for user "{self.user}" on {host}.')
        else:
            inform(f'Will not use a login or password for {host}.')

        # The id's are stored as strings, not ints, to avoid repeated conversion
        self.wanted_list = parsed_id_list(self.id_list)

        if self.report_file:
            if writable(self.report_file):
                inform(f'A report will be written to "{self.report_file}"')
            else:
                alert_fatal(f'Cannot write to file "{self.report_file}"')
                raise CannotProceed(ExitCode.file_error)


    def _do_main_work(self):
        '''Performs the core work of this program.'''
        self._report(f'eprints2archives starting {timestamp()}.', True)

        server = EPrintServer(self.api_url, self.user, self.password)
        available = self._eprints_index(server)
        if not available:
            raise ServerError(f'Received empty list from {server}.')
        self._report(f'EPrints server at {self.api_url} has {len(available)} records.')

        # If the user wants specific records, check which ones actually exist.
        if self.wanted_list:
            missing = sorted(list(set(self.wanted_list) - set(available)), key = int)
            if missing and self.quit_on_error:
                raise ValueError(f'{intcomma(len(missing))} of the requested records'
                                 + ' do not exist on the server: {", ".join(missing)}.')
            elif missing:
                msg = (f"Of the records requested, the following don't exist and"
                       + f" will be skipped: {', '.join(missing)}.")
                warn(msg)
                self._report(msg)
            wanted = sorted(list(set(self.wanted_list) - set(missing)), key = int)
            self._report(f'A total of {len(wanted)} records from {server} will be used.')
        else:
            wanted = available

        # The basic URLs for EPrint pages can be constructed w/o looking up
        # records -- you only need id numbers from the index.  Some sites
        # like Caltech use an additional field, <official_url>, that DOES
        # require a database lookup. We can't know if a site uses it until we
        # try it, and we can't be sure every record has a value, so we have to
        # try it for all records.  If we're not filtering by lastmod or status,
        # it's faster to do direct lookups of <official_url>; conversely, if
        # we DO need lastmod or status values, it takes less time to just get
        # the XML of every record rather than ask for 2 or more field values
        # separately, because each such EPrint lookup is slow and doing 2 or
        # more lookups per record is slower than doing one XML fetch.

        records = []
        if not self.lastmod and not self.status:
            official_url = lambda r: server.eprint_field_value(r, 'official_url')
            ulist = self._eprints_values(official_url, wanted, server, "<official_url>'s")
        else:
            skipped = []
            for r in self._eprints_values(server.eprint_xml, wanted, server, "record materials"):
                eprintid = server.eprint_field_value(r, 'eprintid')
                modtime  = server.eprint_field_value(r, 'lastmod')
                status   = server.eprint_field_value(r, 'eprint_status')
                if self.lastmod and modtime and parsed_datetime(modtime) < self.lastmod:
                    if __debug__: log(f'{eprintid} lastmod == {modtime} -- skipping')
                    skipped.append(r)
                    continue
                if self.status and status and not self._status_acceptable(status):
                    if __debug__: log(f'{eprintid} status == {status} -- skipping')
                    skipped.append(r)
                    continue
                if __debug__: log(f'{eprintid} passed filter checks')
                records.append(r)
            if len(skipped) > 0:
                inform(f'Skipping {len(skipped)} records due to filtering.')
                self._report(f'Skipping {len(skipped)} records due to filtering.')
            if len(records) == 0:
                alert('Filtering left 0 records -- nothing left to do')
                return
            ulist = [server.eprint_field_value(r, 'official_url') for r in records]

        # Filter out invalid URLs from the <official_url> values we gathered.
        if __debug__: log(f'validating list of {len(ulist)} <official_url> URLs')
        urls = []
        for url in ulist:
            if valid_url(url):
                urls.append(url)
            else:
                self._report(f'Ignoring invalid URL: {url}')

        # Next, add "standard" URLs.  Do it AFTER the steps above, b/c if we
        # did any filtering, we may now have a shorter list => less to do.

        urls += self._eprints_record_urls(server, records or wanted)
        urls += self._eprints_general_urls(server, records or wanted)

        # Filter None's & make URLs unique (using a trick with dict.fromkeys).
        if __debug__: log(f'de-duping & validating list of {len(urls)} URLs')
        urls = list(dict.fromkeys(filter(None, urls)))

        # Check parent hasn't raised an interrupt, and if not, start sending.
        raise_for_interrupts()

        self._send(urls)
        inform('Done.')


    def _eprints_values(self, value_function, items_list, server, description):
        '''Get values using "value_function" for items in "items_list".'''
        # Helper function: body of loop that is executed in all cases.
        def record_values(items, update_progress):
            results = []
            for item in items:
                if __debug__: log(f'getting data for record {item}')
                failure = None
                try:
                    data = value_function(item)
                except NoContent as ex:
                    failure = f'Server has no content for {item}'
                except AuthenticationFailure as ex:
                    failure = f'Authentication failure trying to get data for {item}'
                except ServiceFailure as ex:
                    failure = f'{str(ex)} trying to get data for {item}'
                except Exception as ex:
                    raise ex
                if failure:
                    if self.quit_on_error:
                        alert(failure)
                        raise CannotProceed(ExitCode.exception)
                    else:
                        warn(failure)
                        continue
                if data is not None:
                    results.append(data)
                elif self.quit_on_error:
                    alert(f'Received no data for {item}')
                    raise CannotProceed(ExitCode.exception)
                update_progress()
                raise_for_interrupts()
            return results

        server_name = f'[sea_green2]{server}[/]'
        header  = f'[dark_sea_green4]Gathering {description} from {server_name} ...'
        return self._gathered(record_values, items_list, header)


    def _eprints_index(self, server):
        '''Return the index from the server, getting it with a progress bar.'''
        header = f'[dark_sea_green4]Getting full EPrints index from [sea_green2]{server}[/] ...'
        with Progress('[progress.description]{task.description}', _BAR) as progress:
            bar = progress.add_task(header, start = False)
            progress.update(bar)
            index = server.index()
            progress.start_task(bar)
            progress.update(bar, advance = 100)
        return index


    def _eprints_general_urls(self, server, subset):
        '''Return a list of commonly-available, high-level EPrints URLs.

        Parameter value "subset" is taken to be a list of EPrint id's or EPrint
        XML objects that is used to limit the pages under /view to be returned.
        Otherwise, if no "subset" list is given, all /view pages are returned.
        '''
        header = '[dark_sea_green4]Looking through /view pages for URLs ...' + ' '*(len(str(server)) - 4)
        with Progress('[progress.description]{task.description}', _BAR) as progress:
            bar = progress.add_task(header, start = False)
            progress.update(bar)
            if self.wanted_list:
                urls = server.view_urls(subset)
            else:
                urls = unique(server.top_level_urls() + server.view_urls())
            progress.start_task(bar)
            progress.update(bar, advance = 100)
        return urls


    def _eprints_record_urls(self, server, records_list):
        '''Get the normal EPrints URLS for the items in "records_list".'''
        # Helper function: body of loop that is executed in all cases.
        def eprints_urls(item_list, update_progress):
            urls = []
            for r in item_list:
                # Note: don't use log() here b/c r could be an xml etree.
                try:
                    urls.append(server.eprint_id_url(r))
                    urls.append(server.eprint_page_url(r))
                except (NoContent, AuthenticationFailure) as ex:
                    continue
                update_progress()
                raise_for_interrupts()
            return urls

        server_name = f'[sea_green2]{server}[/]'
        header  = f'[dark_sea_green4]Checking variant record URLs on {server_name} ...'
        return self._gathered(eprints_urls, records_list, header)


    def _send(self, urls_to_send):
        '''Send the list of URLs to each web archiving service in parallel.'''
        num_urls = len(urls_to_send)
        num_dest = len(self.dest)

        inform(f'We have a total of {pluralized("URL", num_urls, True)}'
               + f' to send to {pluralized("archive", num_dest, True)}.')
        if self.force:
            inform('Force option given ⟹  adding URLs even if archives have copies.')
        self._report(f'Sending {num_urls} URLs to {pluralized("service", num_dest, True)}.')

        # Helper function: send urls to given service & use progress bar.
        def send_to_service(dest, prog):
            num_added = 0
            num_skipped = 0
            status_text = activity(dest, ServiceStatus.RUNNING)
            row = prog.add_task(status_text, total = num_urls, added = 0, skipped = 0)
            notify = lambda s: prog.update(row, description = activity(dest, s), refresh = True)
            for url in urls_to_send:
                if __debug__: log(f'next for {dest}: {url}')
                (added, num_existing) = dest.save(url, notify, self.force)
                added_str = "added" if added else "skipped"
                num_added += int(added)
                num_skipped += int(not added)
                prog.update(row, advance = 1, added = num_added, skipped = num_skipped)
                self._report(f'{url} ➜ {dest.name}: {added_str}')
                raise_for_interrupts()

        # Start of actual procedure.
        info = TextColumn('{task.fields[added]} added/{task.fields[skipped]} skipped')
        with Progress('[progress.description]{task.description}', _BAR, info) as prog:
            if self.threads == 1:
                # For 1 thread, avoid thread pool to make debugging easier.
                results = [send_to_service(service, prog) for service in self.dest]
            else:
                num_threads = min(num_dest, self.threads)
                if __debug__: log(f'using {num_threads} threads to send records')
                self._executor = ThreadPoolExecutor(max_workers = num_threads,
                                                    thread_name_prefix = 'SendThread')
                self._futures = []
                for service in self.dest:
                    future = self._executor.submit(send_to_service, service, prog)
                    self._futures.append(future)
                [f.result() for f in self._futures]
        self._report(f'Finished sending {num_urls} URLs.')


    def _status_acceptable(self, this_status):
        '''Return True if "this_status" should be accepted based on filters.'''
        # The presence of '^' indicates negation, i.e., "not any of these".
        return ((not '^' in self.status and this_status in self.status)
                or ('^' in self.status and this_status not in self.status))


    def _gathered(self, loop, items_list, header):
        '''Return the collected results of running "loop" in multiple threads.'''
        num_items = len(items_list)
        text = TextColumn('{task.completed}/' + intcomma(num_items) + ' records')
        with Progress('[progress.description]{task.description}', _BAR, text) as progress:
            # Wrap up the progress bar updater as a lambda that we pass down.
            bar = progress.add_task(header, total = num_items)
            update_progress = lambda: progress.update(bar, advance = 1)

            # If the number of items is too small, don't bother going parallel.
            if self.threads == 1 or num_items <= (self.threads * _PARALLEL_THRESHOLD):
                return loop(items_list, update_progress)

            # If we didn't return above, we're going parallel.
            num_threads = min(num_items, self.threads)
            if __debug__: log(f'using {num_threads} threads to gather records')
            self._executor = ThreadPoolExecutor(max_workers = num_threads,
                                                thread_name_prefix = 'GatherThread')
            self._futures = []
            for sublist in slice(items_list, num_threads):
                future = self._executor.submit(loop, sublist, update_progress)
                self._futures.append(future)
            # We get a list of lists, so flatten it before returning.
            return flatten(f.result() for f in self._futures)


    def _report(self, text, overwrite = False):
        '''Write text to the report file, if a report file is being written.'''
        # Opening/closing the file for every write is inefficient, but our
        # network operations are slow, so I think this is not going to have
        # enough impact to be concerned
        if __debug__: log(text)
        if self.report_file:
            with open(self.report_file, 'w' if overwrite else 'a') as f:
                f.write(text + os.linesep)

# Helper functions.
# ......................................................................

def parsed_id_list(id_list):
    if id_list is None:
        return []

    # If it's a single digit, asssume it's not a file and return the number.
    if id_list.isdigit():
        return [id_list]

    # Things get trickier because anything else could be (however improbably)
    # a file name.  So use a process of elimination: try to see if a file by
    # that name exists, and if it doesn't, parse the argument as numbers.
    candidate = id_list
    if not path.isabs(candidate):
        candidate = path.realpath(path.join(os.getcwd(), candidate))
    if path.exists(candidate):
        if not readable(candidate):
            raise RuntimeError(f'File not readable: {candidate}')
        with open(candidate, 'r', encoding = 'utf-8-sig') as file:
            if __debug__: log(f'reading {candidate}')
            return [id.strip() for id in file.readlines()]

    # Didn't find a file.  Try to parse as multiple numbers.
    if ',' not in id_list and '-' not in id_list:
        raise ValueError('Unable to understand list of record identifiers')
    return list(flatten(expanded_range(x) for x in id_list.split(',')))


def fmt_statuses(status_list, negated):
    as_list = ['"' + x + '"' for x in status_list]
    if len(as_list) > 1:
        and_or = ' or ' if negated else ' and '
        return ', '.join(as_list[:-1]) + and_or + as_list[-1]
    else:
        return as_list[0]


def activity(dest, status):
    name = f'[{dest.color}]{dest.name}[/]'
    if status == ServiceStatus.RUNNING:
        return f'[dark_sea_green4]Sending URLs to {name} ...                     '
    elif status == ServiceStatus.PAUSED_RATE_LIMIT:
        return f'[yellow3 on grey35]Paused for rate limit {name} ...               '
    elif status == ServiceStatus.PAUSED_ERROR:
        return f'[orange1]Paused due to {name} error -- will retry ...   '
    elif status == ServiceStatus.UNAVAILABLE:
        return f'[red]No response from {name} servers ...            '
    else:
        return f'[red]Unknown status                                 '



'''
eprints2archives interfaces to web archiving services.

Authors
-------

Michael Hucka <mhucka@caltech.edu> -- Caltech Library

Copyright
---------

Copyright (c) 2020-2021 by the California Institute of Technology.  This code
is open-source software released under a 3-clause BSD license.  Please see the
file "LICENSE" for more information.
'''

from enum import Enum

from .upload_status import ServiceStatus
from .internetarchive import InternetArchive
from .archivetoday import ArchiveToday


# Constants.
# .............................................................................

KNOWN_SERVICES = {
    'internetarchive' : InternetArchive(),
    'archivetoday'    : ArchiveToday(),
}

# Save this list to avoid recreating it all the time.
SERVICE_NAMES = sorted(KNOWN_SERVICES.keys())
SERVICE_OBJECTS = KNOWN_SERVICES.values()


# Exported functions.
# .............................................................................

def service_names():
    '''Return a list of the known service names.'''
    return SERVICE_NAMES


def service_interfaces():
    '''Return a list of objects that act as interfaces to services.'''
    return SERVICE_OBJECTS


def service_by_name(name):
    '''Return the object corresponding to the given service "name".

    If "name" is is not a known service name, the value None is returned.
    '''
    return KNOWN_SERVICES.get(name.lower(), None)



'''
eprints.py: EPrints-specific utilities

Authors
-------

Michael Hucka <mhucka@caltech.edu> -- Caltech Library

Copyright
---------

Copyright (c) 2020-2021 by the California Institute of Technology.  This code
is open-source software released under a 3-clause BSD license.  Please see the
file "LICENSE" for more information.
'''

from   bun import alert, alert_fatal, warn
import codecs
from   collections import defaultdict
from   commonpy.data_utils import parsed_datetime, unique
from   commonpy.network_utils import net, hostname, scheme, netloc
from   commonpy.exceptions import NoContent, AuthenticationFailure, RateLimitExceeded
import httpx
from   lxml import etree, html
import os
from   os import path
import shutil
import ssl
import validators

if __debug__:
    from sidetrack import log

from .exceptions import *


# Constants.
# .............................................................................

_EPRINTS_XMLNS = 'http://eprints.org/ep2/data/2.0'
'''XML namespace used in EPrints XML output.'''


# Main functions.
# .............................................................................

class EPrintServer():

    def __init__(self, given_url, user, password):
        if __debug__: log(f'creating EPrintsServer object for {given_url}')
        # For efficiency, create an httpx Client object and reuse it (since
        # we're always talking to the same EPrints server).  Note about SSL
        # security settings: EPrints, being an older system, is often run on
        # older servers with older SSL support. The following enables SSL3.0
        # support, which I found is needed for EPrints servers that I tried.
        # C.f. https://docs.python.org/3/library/ssl.html#ssl.SSLContext
        ssl_config = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
        ssl_config.options &= ~ssl.OP_NO_SSLv3
        timeout = httpx.Timeout(30, connect = 30, read = 30, write = 30)
        self._client = httpx.Client(timeout = timeout, http2 = True, verify = ssl_config)

        # Do this after the above b/c it needs the network client to be set up.
        self._api_url    = self._canonical_endpoint_url(given_url)

        self._protocol   = scheme(self._api_url)
        self._netloc     = netloc(self._api_url)
        self._hostname   = hostname(self._api_url)
        self._base_url   = self._protocol + '://' + self._netloc
        self._user       = user
        self._password   = password
        # List of all record identifiers known to the server:
        self._index      = []
        # Cache of record XML objects obtained from the server:
        self._records    = {}


    def __str__(self):
        return self._hostname


    # Public methods ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    def api_url(self):
        '''Return the canonical REST API URL for this server.'''
        return self._api_url


    def index(self, as_int = False):
        '''Return a list of all record identifiers from the server.  By
        default, the numbers are returned as strings, not ints, because the
        most common uses of the results end up using strings anyway.  However,
        if the parameter 'as_int' is set to True, the list will be returned
        as integers.'''

        # The raw index returned by EPrints is in XHTML format, like this:
        #
        #   <!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
        #       "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
        #   <html xmlns="http://www.w3.org/1999/xhtml">
        #   <head>
        #     <title>EPrints REST: Eprints DataSet</title>
        #     <style type="text/css">
        #       body { font-family: sans-serif; }
        #     </style>
        #   </head>
        #   <body>
        #     <h1>EPrints REST: Eprints DataSet</h1>
        #   <ul>
        #   <li><a href='4/'>4/</a></li>
        #   <li><a href='4.xml'>4.xml</a></li>
        #   <li><a href='5/'>5/</a></li>
        #   <li><a href='5.xml'>5.xml</a></li>
        #   ...
        # We extract the identifiers by looping over the XML elements.

        if self._index:
            if __debug__: log('using cached copy of index')
        else:
            xml = etree.fromstring(self._eprints_raw_index())
            if __debug__: log("parsing raw EPrints index to get record id's")
            for node in xml.findall('.//{http://www.w3.org/1999/xhtml}a'):
                if 'href' in node.attrib and node.attrib['href'].endswith('xml'):
                    self._index.append(node.attrib['href'].split('.')[0])
        if self._index and as_int:
            if __debug__: log('returning list of identifiers as ints')
            return [int(x) for x in self._index]
        else:
            return self._index


    def front_page_url(self):
        '''Return the public front page URL of this EPrints server.'''
        return self._base_url


    def top_level_urls(self):
        '''Return a list of the top-level URLs of this EPrints server.

        This returns all URLs found by examining the HTML of the front page
        of the server and then keeping only those located on the server and
        not certain special URLs like CGI, css, relative links, and so on.
        '''
        top_page = self._base_url
        if __debug__: log(f'getting from page from {self._base_url}')
        (response, error) = self._net('get', top_page)
        if error:
            if __debug__: log(f'got {type(error)} error for {top_page}')
            return []
        # Scrape the HTML.
        doc = html.fromstring(response.text)
        doc.make_links_absolute(top_page)
        # Extract unique URLs, filtering out stuff we don't want.
        skip = ['/cgi', '#', 'css']
        keep = lambda u: u and u.startswith(top_page) and not any(x in u for x in skip)
        urls = unique(filter(keep, [x.get('href') for x in doc.cssselect('a')]))
        if __debug__: log(f'found {len(urls)} top-level URLs: {urls}')
        return urls


    def view_urls(self, subset = None):
        '''Return a list of URLs corresponding to pages under /view.

        If parameter value "subset" is not None, it is taken to be a list
        of EPrint id's or XML objects, and used to limit the pages under to be
        returned.  The values return will only consist of URLs of the form
        /view/X/N, where X is a subpage under /view and N is the id of a
        record in "subset", if such URLs exist on the server.  Otherwise,
        if no "subset" list is given, all /view/ids/N pages are returned
        (again, if they exist), along with all other pages found under /view
        and the pages one level below each of those. (E.g., /view/year,
        /view/year/yyyy.html, /view/person-az, /view/person-az/a.html, etc.)
        '''
        # We need /view in any case.
        view_base = self._base_url + '/view/'
        (response, error) = self._net('get', view_base)
        if error:
            if __debug__: log(f'got {type(error)} error for {view_base}')
            return []

        # Scrape the HTML to find the block of links to pages under /view.
        if __debug__: log(f'scraping HTML of {view_base}')
        doc = html.fromstring(response.text)
        doc.make_links_absolute(view_base)
        view_urls = set(x.get('href') for x in doc.cssselect('div.ep_view_browse_list li a'))
        if __debug__: log(f'found {len(view_urls)} URLs under /view')

        # Iterate over each page under /view, to get links to their subpages.
        subpage_urls = set()
        for subpage in view_urls:
            (response, error) = self._net('get', subpage)
            if error:
                if __debug__: log(f'got {type(error)} error for {subpage}')
                continue
            if __debug__: log(f'scraping HTML of {subpage}')
            doc = html.fromstring(response.text)
            doc.make_links_absolute(subpage)
            subpage_urls |= set(x.get('href') for x in doc.cssselect('div.ep_view_menu li a'))
        if __debug__: log(f'collected {len(subpage_urls)} /view subpage URLs')

        # If subset is given, we ONLY keep pages of the form /view/X/N.html
        if subset:
            kept_urls = set()
            for item in subset:
                for url in subpage_urls:
                    if isinstance(item, (str, int)):
                        eprintid = item
                    else:
                        eprintid = self._xml_field_value(item, 'eprintid')
                    # Year pages will have the form N.html too.  Skip them.
                    if '/view/year' not in url and url.endswith(f'/{eprintid}.html'):
                        if __debug__: log(f'keeping {eprintid}.html')
                        kept_urls.add(url)
                        break
            if __debug__: log(f'returning subset {len(kept_urls)} /view/X/N.html URLs')
            return list(kept_urls)
        else:
            # No subset, so we return everything.
            view_urls |= subpage_urls
            if __debug__: log(f'returning {len(view_urls)} /view subpage URLs')
            return list(view_urls)


    def eprint_id_url(self, id_or_record, verify = True):
        '''Return the main URL of the web page for the record on this server.

        This URL is for the page located at "https://.../id/eprint/N", where N
        is the eprint id of the record in question.  Note that EPrints servers
        typically also make the same records available from a URL of the form
        "https://.../N"; this other form is returned by the companion method
        eprint_page_url().

        The argument value for "id_or_record" can be either an identifier or
        an EPrint XML object (as a Python lxml.etree object).

        A server's index may report EPrint record identifiers for records that
        are not actually exposed to the public (e.g., because they've been
        deleted or otherwise marked as not publicly displayed).  Thus, given
        an identifier or even an XML record, it is possible to create URLs
        that do not lead to valid web pages.  Without additional knowledge
        about a specific server's conventions for public/private record
        visibility, the only way to know is to attempt to access the web page
        at the expected address.  Consequently, this method will attempt to do
        a network lookup on the URL, unless the parameter "verify" is given a
        value of False.
        '''
        if isinstance(id_or_record, (str, int)):
            url = f'{self._protocol}://{self._netloc}/id/eprint/{id_or_record}'
        else:
            eprintid = self._xml_field_value(id_or_record, 'eprintid')
            url = f'{self._protocol}://{self._netloc}/id/eprint/{eprintid}'
        if verify:
            (response, error) = self._net('head', url)
            if error:
                if __debug__: log(f'failed to get /id from server for {id_or_record}')
                return None
        return url


    def eprint_page_url(self, id_or_record, verify = True):
        '''Return the web page address for the record on this server.

        This URL is for the page located at "https://.../N", where N is the
        eprint id of the record in question.  Note that EPrints servers
        typically also make the same records available from a URL of the form
        "https://.../id/eprint/N"; this other form is returned by the companion
        method eprint_id_url().

        The argument value for "id_or_record" can be either an identifier or
        an EPrint XML object (as a Python lxml.etree object).

        A server's index may report EPrint record identifiers for records that
        are not actually exposed to the public (e.g., because they've been
        deleted or otherwise marked as not publicly displayed).  Thus, given
        an identifier or even an XML record, it is possible to create URLs
        that do not lead to valid web pages.  Without additional knowledge
        about a specific server's conventions for public/private record
        visibility, the only way to know is to attempt to access the web page
        at the expected address.  Consequently, this method will attempt to do
        a network lookup on the URL, unless the parameter "verify" is given a
        value of False.
        '''
        if isinstance(id_or_record, (str, int)):
            url = f'{self._protocol}://{self._netloc}/{id_or_record}'
        else:
            eprintid = self._xml_field_value(id_or_record, 'eprintid')
            url = f'{self._protocol}://{self._netloc}/{eprintid}'
        if verify:
            (response, error) = self._net('head', url)
            if error:
                if __debug__: log(f'failed to get /id from server for {id_or_record}')
                return None
        return url


    def eprint_xml(self, eprintid):
        '''Return an XML object identified by the given record identifier.'''
        eprintid = str(eprintid)
        if eprintid in self._records:
            if __debug__: log(f'returning cached XML for record {eprintid}')
            return self._records[eprintid]

        if __debug__: log(f'getting XML for {eprintid} from server')
        try:
            response = self._get_authenticated(f'/eprint/{eprintid}.xml')
        except NoContent as ex:
            if __debug__: log(f'No content for /eprint/{eprintid}.xml')
            raise
        except AuthenticationFailure as ex:
            if __debug__: log(f'Auth failure for /eprint/{eprintid}.xml')
            raise
        except Exception as ex:
            # Our EPrints server sometimes returns with access forbidden for
            # specific records.  Our caller may simply move on, so we store
            # a value before bubbling up the exception.
            if __debug__: log(f'{str(ex)} for {eprintid}')
            self._records[eprintid] = None
            raise ex
        if response is None:
            self._records[eprintid] = None
            return None
        xml = etree.fromstring(response.content)
        self._records[eprintid] = xml
        return xml


    def eprint_field_value(self, id_or_record, field):
        '''Return the value of 'field' of the record 'id_or_record'.  If
        'id_or_record' is a number (either as a string or an integer), this
        method will ask the server for the field value using the REST API;
        if 'id_or_record' is an XML object, then the value is looked up in the
        object itself.'''

        if isinstance(id_or_record, (str, int)):
            id_or_record = str(id_or_record)
            if field == 'eprintid':
                # Data lookups are unnecessary since we have the id already.
                # This case is here to provide a uniform calling experience.
                return id_or_record
            elif id_or_record in self._records:
                # We have a copy of the XML for this one.  Use it.
                if __debug__: log(f'using cached copy of record {id_or_record}')
                xml = self._records[id_or_record]
            else:
                # Contact the server.
                if __debug__: log(f'{id_or_record} not cached -- asking server')
                field_url = f'/eprint/{id_or_record}/{field}.txt'
                try:
                    response = self._get_authenticated(field_url)
                except NoContent as ex:
                    if __debug__: log(f'No content for {field} in {id_or_record}')
                    raise
                except AuthenticationFailure as ex:
                    if __debug__: log(f'Auth failure for {field} in {id_or_record}')
                    raise
                except Exception as ex:
                    if __debug__: log(f'{str(ex)} for {field} in {id_or_record}')
                    raise
                if __debug__: log(f'got response: {response.text}')
                return response.text if response and response.text != '' else None
        else:
            xml = id_or_record
        value = self._xml_field_value(xml, field)
        if __debug__: log(f'obtained value: {value}')
        return value


    # Internal methods ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    def _net(self, method, url):
        '''Make a network call using our preconfigured client.'''
        return net(method, url, client = self._client)


    def _canonical_endpoint_url(self, url):
        '''Return the canonical REST API URL.'''
        # We remove any ending /eprint because we add it separately when needed.
        # For convenience, we add /rest if the user forgot.  Ditto for https://.
        if not scheme(url):
            # Is the server at https://, or http://?
            if __debug__: log(f'trying to add https or http to {url}')
            for prefix in ['https://', 'http://']:
                candidate = prefix + url
                try:
                    (response, error) = self._net('head', candidate)
                except:
                    continue
                url = candidate
                break
        if url.endswith('/'):
            url = url[0 : -1]
        if url.endswith('/eprint'):
            url = url[0 : url.rfind('/eprint')]
        if not url.endswith('/rest'):
            url += '/rest'
        if not validators.url(url):
            alert_fatal(f'The given API URL appears invalid: {url}')
            raise CannotProceed(ExitCode.bad_arg)
        return url


    def _get_authenticated(self, op):
        if not self._api_url:
            return None

        url = self._api_url
        start = url.find('//')
        if start < 0:
            raise ValueError(f'Unable to parse "{url}" as a normal URL')
        if self._user and self._password:
            endpoint = url[:start + 2] + self._user + ':' + self._password + '@' + url[start + 2:] + op
        elif self._user and not self._password:
            endpoint = url[:start + 2] + self._user + '@' + url[start + 2:] + op
        else:
            endpoint = url[:start + 2] + url[start + 2:] + op

        (response, error) = self._net('get', endpoint)
        if response and not error:
            return response
        else:
            if __debug__: log(f'got {type(error)} error for {url}')
            raise error


    def _eprints_raw_index(self):
        if __debug__: log(f'asking {self._hostname} for index of records')
        response = self._get_authenticated('/eprint')
        if response and response.text and response.text.startswith('<?xml'):
            return response.content
        else:
            # This shouldn't happen.
            raise InternalError('Internal error processing server response')


    def _xml_field_value(self, xml, field):
        if xml is not None and etree.iselement(xml):
            node = xml.find('.//{' + _EPRINTS_XMLNS + '}' + field)
            return node.text if node != None else ''
        else:
            raise InternalError(f'Not an XML object: {xml}')
